<?php
/**
 * Table Definition for tags
 *
 * PHP version 7
 *
 * Copyright (C) Villanova University 2010.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  Db_Table
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
namespace fid_bbi\Db\Table;

use VuFind\Db\Table\Tags as BaseTags;
use Zend\Db\Sql\Expression;
use Zend\Db\Sql\Select;

/**
 * Table Definition for tags
 *
 * @category VuFind
 * @package  Db_Table
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
class Tags extends BaseTags
{
    use ResourceTagsTrait;

    /**
     * Get a list of all tags generated by the user in favorites lists.  Note that
     * the returned list WILL NOT include tags attached to records that are not
     * saved in favorites lists.
     *
     * @param string $userId     User ID to look up.
     * @param string $resourceId Filter for tags tied to a specific resource (null
     * for no filter).
     * @param int    $listId     Filter for tags tied to a specific list (null for no
     * filter).
     * @param string $source     Filter for tags tied to a specific record source
     * (null for no filter).
     *
     * @return \Zend\Db\ResultSet\AbstractResultSet
     */
    public function getForUser($userId, $resourceId = null, $listId = null,
        $source = null
    ) {
        $callback = function ($select) use ($userId, $resourceId, $listId, $source) {
            $select->columns(
                [
                    'id' => new Expression(
                        'min(?)', ['tags.id'],
                        [Expression::TYPE_IDENTIFIER]
                    ),
                    'tag' => $this->caseSensitive
                        ? 'tag' : new Expression('lower(tag)'),
                    'cnt' => new Expression(
                        'COUNT(DISTINCT(?))', ['rt.resource_id'],
                        [Expression::TYPE_IDENTIFIER]
                    )
                ]
            );
            $select->join(
                ['rt' => 'resource_tags'], 'tags.id = rt.tag_id', []
            );
            $select->join(
                ['r' => 'resource'], 'rt.resource_id = r.id', []
            );
            $select->join(
                ['ur' => 'user_resource'], 'r.id = ur.resource_id', []
            );
            $select->group(['tag'])->order([new Expression('lower(tag)')]);

            // contrary to the VuFind core approach we want all
            // TAGS of RECORDS from the list
            // not only those tagged within the list
            $select->where->equalTo('rt.user_id', $userId);

            if (null !== $source) {
                $select->where->equalTo('r.source', $source);
            }

            if (null !== $resourceId) {
                $select->where->equalTo('r.record_id', $resourceId);
            }
            if (null !== $listId) {
                // same as above: all TAGS of RECORDS from the list
                $select->where->equalTo('ur.list_id', $listId);
            }
        };
        return $this->select($callback);
    }

    /**
     * Get a list of all tags generated by the user in favorites lists.  Note that
     * the returned list WILL NOT include tags attached to records that are not
     * saved in favorites lists.
     *
     * @param string $userId     User ID to look up.
     * @param string $resourceId Filter for tags tied to a specific resource (null
     * for no filter).
     * @param int    $listId     Filter for tags tied to a specific list (null for no
     * filter).
     * @param string $source     Filter for tags tied to a specific record source
     * (null for no filter).
     *
     * @return \Zend\Db\ResultSet\AbstractResultSet
     */
    public function getAllForUser($userId, $source = null)
    {
        $callback = function ($select) use ($userId, $source) {
            $select->columns(
                [
                    'id' => new Expression(
                        'min(?)', ['tags.id'],
                        [Expression::TYPE_IDENTIFIER]
                    ),
                    'tag' => $this->caseSensitive
                        ? 'tag' : new Expression('lower(tag)'),
                    'cnt' => new Expression(
                        'COUNT(DISTINCT(?))', ['rt.resource_id'],
                        [Expression::TYPE_IDENTIFIER]
                    ),
                    'lists' => new Expression(
                        'GROUP_CONCAT(DISTINCT(?))', ['ur.list_id'],
                        [Expression::TYPE_IDENTIFIER]
                    )
                ]
            );
            $select->join(
                ['rt' => 'resource_tags'], 'tags.id = rt.tag_id', []
            );
            $select->join(
                ['r' => 'resource'], 'rt.resource_id = r.id', []
            );
            $select->join(
                ['ur' => 'user_resource'], 'r.id = ur.resource_id', [], \Zend\Db\Sql\Select::JOIN_LEFT_OUTER
            );
            $select->group(['tag'])->order([new Expression('lower(tag)')]);

            // contrary to the VuFind core approach we want all
            // TAGS of RECORDS from the list
            // not only those tagged within the list
            $select->where->equalTo('rt.user_id', $userId);

            if (null !== $source) {
                $select->where->equalTo('r.source', $source);
            }
        };
        return $this->select($callback);
    }

    /**
     * Get tags associated with the specified resource.
     *
     * @param string $id          Record ID to look up
     * @param string $source      Source of record to look up
     * @param int    $limit       Max. number of tags to return (0 = no limit)
     * @param int    $list        ID of list to load tags from (null for no
     * restriction,  true for on ANY list, false for on NO list)
     * @param int    $user        ID of user to load tags from (null for all users)
     * @param string $sort        Sort type ('count' or 'tag')
     * @param int    $userToCheck ID of user to check for ownership (this will
     * not filter the result list, but rows owned by this user will have an is_me
     * column set to 1)
     *
     * @return Iterable
     */
    public function getForResource($id, $source = DEFAULT_SEARCH_BACKEND, $limit = 0,
                                   $list = null, $user = null, $sort = 'count', $userToCheck = null
    ) {
        return $this->select(
            function ($select) use (
                $id, $source, $limit, $list, $user, $sort, $userToCheck
            ) {
                // If we're looking for ownership, create sub query to merge in
                // an "is_me" flag value if the selected resource is tagged by
                // the specified user.
                if (!empty($userToCheck)) {
                    $subq = $this->getIsMeSubquery($id, $source, $userToCheck);
                    $select->join(
                        ['subq' => $subq],
                        'tags.id = subq.tag_id',
                        [
                            // is_me will either be null (not owned) or the ID
                            // of the tag (owned by the current user).
                            'is_me' => new Expression(
                                'MAX(?)', ['subq.tag_id'],
                                [Expression::TYPE_IDENTIFIER]
                            )
                        ],
                        Select::JOIN_LEFT
                    );
                }
                // SELECT (do not add table prefixes)
                $select->columns(
                    [
                        'id',
                        'tag' => $this->caseSensitive
                            ? 'tag' : new Expression('lower(tag)'),
                        'cnt' => new Expression(
                            'COUNT(DISTINCT(?))', ["rt.user_id"],
                            [Expression::TYPE_IDENTIFIER]
                        )
                    ]
                );
                $select->join(
                    ['rt' => 'resource_tags'], 'rt.tag_id = tags.id', []
                );
                $select->join(
                    ['r' => 'resource'], 'rt.resource_id = r.id', []
                );
                $select->where(['r.record_id' => $id, 'r.source' => $source]);
                $select->group(['tags.id', 'tag']);

                if ($sort == 'count') {
                    $select->order(['cnt DESC', new Expression('lower(tags.tag)')]);
                } elseif ($sort == 'tag') {
                    $select->order([new Expression('lower(tags.tag)')]);
                }

                if ($limit > 0) {
                    $select->limit($limit);
                }
                if ($list === true) {
                    $select->where->isNotNull('rt.list_id');
                } elseif ($list === false) {
                    $select->where->isNull('rt.list_id');
                } elseif (null !== $list) {
                    $select->where->equalTo('rt.list_id', $list);
                }
                if (null !== $user) {
                    $select->where->equalTo('rt.user_id', $user);
                }
                $this->addFirstPosted($select, $userToCheck);
            }
        );
    }
}
